__version__ = "0.0.1"


#from ._widget import PalmariWidget
from .data_structure.experiment import Experiment
from .data_structure.acquisition import Acquisition
from .processing.image_pipeline import ImagePipeline

# from .processing.image_pipeline_widget import ImagePipelineWidget
